/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "DebugDrawerHelper.h"

#include <VirtualRobot/math/Helpers.h>

#include <ArmarXCore/util/CPPUtility/Iterator.h>

#include <RobotAPI/libraries/core/Pose.h>

using namespace math;
using namespace armarx;

#define CHECK_AND_ADD(name,type) if (!enableVisu) {return;} addNewElement(name,type);
//#define CHECK_AND_ADD(name,type) if (!enableVisu) {return;}

DebugDrawerHelper::DebugDrawerHelper(const DebugDrawerInterfacePrx& debugDrawerPrx, const std::string& layerName, const VirtualRobot::RobotPtr& robot)
    : debugDrawerPrx(debugDrawerPrx), layerName(layerName), robot(robot), rn(robot ? robot->getRootNode() : VirtualRobot::RobotNodePtr())
{
}

//1st order
void DebugDrawerHelper::drawPose(const std::string& name, const Eigen::Matrix4f& pose)
{
    CHECK_AND_ADD(name, DrawElementType::Pose)
    debugDrawerPrx->setPoseVisu(layerName, name, makeGlobal(pose));
}

void DebugDrawerHelper::drawPose(const std::string& name, const Eigen::Matrix4f& pose, float scale)
{
    CHECK_AND_ADD(name, DrawElementType::Pose)
    debugDrawerPrx->setScaledPoseVisu(layerName, name, makeGlobal(pose), scale);
}

void DebugDrawerHelper::drawBox(const std::string& name, const Eigen::Vector3f& position, float size, const DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::Box)
    drawBox(name, Helpers::CreatePose(position, Eigen::Matrix3f::Identity()), Eigen::Vector3f(size, size, size), color);
}

void DebugDrawerHelper::drawBox(const std::string& name, const Eigen::Matrix4f& pose, float size, const DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::Box)
    drawBox(name, pose, Eigen::Vector3f(size, size, size), color);
}

void DebugDrawerHelper::drawBox(const std::string& name, const Eigen::Matrix4f& pose, const Eigen::Vector3f& size, const DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::Box)
    debugDrawerPrx->setBoxVisu(layerName, name, makeGlobal(pose), new Vector3(size), color);
}

void DebugDrawerHelper::drawLine(const std::string& name, const Eigen::Vector3f& p1, const Eigen::Vector3f& p2, float width, const DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::Line)
    debugDrawerPrx->setLineVisu(layerName, name, makeGlobal(p1), makeGlobal(p2), width, color);
}

void DebugDrawerHelper::drawText(const std::string& name, const Eigen::Vector3f& p1, const std::string& text, const DrawColor& color, int size)
{
    CHECK_AND_ADD(name, DrawElementType::Text)
    debugDrawerPrx->setTextVisu(layerName, name, text, makeGlobal(p1), color, size);
}

void DebugDrawerHelper::drawArrow(const std::string& name, const Eigen::Vector3f& pos, const Eigen::Vector3f& direction, const DrawColor& color, float length, float width)
{
    CHECK_AND_ADD(name, DrawElementType::Arrow)
    debugDrawerPrx->setArrowVisu(layerName, name, makeGlobal(pos), makeGlobalDirection(direction), color, length, width);
}

void DebugDrawerHelper::drawPointCloud(const std::string& name, const std::vector<Eigen::Vector3f>& points, float pointSize, const DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::ColoredPointCloud)
    DebugDrawerColoredPointCloud pc;
    pc.pointSize = pointSize;
    for (const Eigen::Vector3f& p : points)
    {
        Eigen::Vector3f pg = rn ? rn->getGlobalPosition(p) : p;
        DebugDrawerColoredPointCloudElement e;
        e.x = pg(0);
        e.y = pg(1);
        e.z = pg(2);
        e.color = color; //DrawColor24Bit {(Ice::Byte)(color.r * 255), (Ice::Byte)(color.g * 255), (Ice::Byte)(color.b * 255)};
        pc.points.push_back(e);
    }
    debugDrawerPrx->setColoredPointCloudVisu(layerName, name, pc);
}

void armarx::DebugDrawerHelper::drawRobot(const std::string& name, const std::string& robotFile, const std::string& armarxProject, const Eigen::Matrix4f& pose, const armarx::DrawColor& color)
{
    CHECK_AND_ADD(name, DrawElementType::Robot)
    debugDrawerPrx->setRobotVisu(layerName, name, robotFile, armarxProject, DrawStyle::FullModel);
    debugDrawerPrx->updateRobotPose(layerName, name, makeGlobal(pose));
    debugDrawerPrx->updateRobotColor(layerName, name, color);
}

void DebugDrawerHelper::setRobotConfig(const std::string& name, const std::map<std::string, float>& config)
{
    debugDrawerPrx->updateRobotConfig(layerName, name, config);
}

//2nd order
void DebugDrawerHelper::drawPoses(const std::string& prefix, const std::vector<Eigen::Matrix4f>& poses)
{
    for (const auto& [idx, pose] : MakeIndexedContainer(poses))
    {
        drawPose(prefix + std::to_string(idx), pose);
    }
}
void DebugDrawerHelper::drawPoses(const std::string& prefix, const std::vector<Eigen::Matrix4f>& poses, float scale)
{
    for (const auto& [idx, pose] : MakeIndexedContainer(poses))
    {
        drawPose(prefix + std::to_string(idx), pose, scale);
    }
}

void DebugDrawerHelper::drawLine(const std::string& name, const Eigen::Vector3f& p1, const Eigen::Vector3f& p2)
{
    drawLine(name, p1, p2, defaults.lineWidth, defaults.lineColor);
}

void DebugDrawerHelper::drawLines(const std::string& prefix, const std::vector<Eigen::Vector3f>& ps, float width, const DrawColor& color)
{
    for (std::size_t idx = 1; idx < ps.size(); ++idx)
    {
        drawLine(prefix + std::to_string(idx), ps.at(idx - 1), ps.at(idx), width, color);
    }
}
void DebugDrawerHelper::drawLines(const std::string& prefix, const std::vector<Eigen::Vector3f>& ps)
{
    for (std::size_t idx = 1; idx < ps.size(); ++idx)
    {
        drawLine(prefix + std::to_string(idx), ps.at(idx - 1), ps.at(idx));
    }
}
void DebugDrawerHelper::drawLines(const std::string& prefix, const std::vector<Eigen::Matrix4f>& ps, float width, const DrawColor& color)
{
    for (std::size_t idx = 1; idx < ps.size(); ++idx)
    {
        drawLine(prefix + std::to_string(idx),
                 ps.at(idx - 1).topRightCorner<3, 1>(),
                 ps.at(idx).topRightCorner<3, 1>(),
                 width, color);
    }
}
void DebugDrawerHelper::drawLines(const std::string& prefix, const std::vector<Eigen::Matrix4f>& ps)
{
    for (std::size_t idx = 1; idx < ps.size(); ++idx)
    {
        drawLine(prefix + std::to_string(idx),
                 ps.at(idx - 1).topRightCorner<3, 1>(),
                 ps.at(idx).topRightCorner<3, 1>());
    }
}

//utility
void DebugDrawerHelper::clearLayer()
{
    debugDrawerPrx->clearLayer(layerName);
    oldElements.clear();
    newElements.clear();
}

void DebugDrawerHelper::cyclicCleanup()
{
    /*
    // debug code...
    std::stringstream ss;
    ss << "OLD: ";
    for (const DrawElement& e : oldElements)
    {
        ss << e.name << ",";
    }
    ss << " NEW: ";
    for (const DrawElement& e : newElements)
    {
        ss << e.name << ",";
    }
    ARMARX_IMPORTANT << ss.str();*/

    for (const DrawElement& old : oldElements)
    {
        if (newElements.find(old) == newElements.end())
        {
            //ARMARX_IMPORTANT << "removing " << old.name;
            removeElement(old.name, old.type);
        }
    }
    oldElements = newElements;
    newElements.clear();
}

void DebugDrawerHelper::setVisuEnabled(bool enableVisu)
{
    this->enableVisu = enableVisu;
}

PosePtr DebugDrawerHelper::makeGlobal(const Eigen::Matrix4f& pose)
{
    Eigen::Matrix4f gp = rn ? rn->getGlobalPose(pose) : pose;
    return new Pose(gp);
}

Vector3Ptr DebugDrawerHelper::makeGlobal(const Eigen::Vector3f& position)
{
    Eigen::Vector3f gp = rn ? rn->getGlobalPosition(position) : position;
    return new Vector3(gp);
}

Vector3Ptr DebugDrawerHelper::makeGlobalDirection(const Eigen::Vector3f& direction)
{
    Eigen::Vector3f gd = rn ? math::Helpers::TransformDirection(rn->getGlobalPose(), direction) : direction;
    return new Vector3(gd);
}

Eigen::Vector3f DebugDrawerHelper::makeGlobalEigen(const Eigen::Vector3f& position)
{
    return rn ? rn->getGlobalPosition(position) : position;
}

void armarx::DebugDrawerHelper::drawSphere(const std::string& name, const Eigen::Vector3f& position, float size, const armarx::DrawColor& color)
{
    debugDrawerPrx->setSphereVisu(layerName, name, makeGlobal(position), color, size);
}

void DebugDrawerHelper::removeElement(const std::string& name, DebugDrawerHelper::DrawElementType type)
{
    switch (type)
    {
        case DrawElementType::Pose:
            debugDrawerPrx->removePoseVisu(layerName, name);
            return;
        case DrawElementType::Box:
            debugDrawerPrx->removeBoxVisu(layerName, name);
            return;
        case DrawElementType::Line:
            debugDrawerPrx->removeLineVisu(layerName, name);
            return;
        case DrawElementType::Text:
            debugDrawerPrx->removeTextVisu(layerName, name);
            return;
        case DrawElementType::Arrow:
            debugDrawerPrx->removeArrowVisu(layerName, name);
            return;
        case DrawElementType::ColoredPointCloud:
            debugDrawerPrx->removeColoredPointCloudVisu(layerName, name);
            return;
        case DrawElementType::Robot:
            debugDrawerPrx->removeRobotVisu(layerName, name);
            return;
    }
}

void DebugDrawerHelper::addNewElement(const std::string& name, DebugDrawerHelper::DrawElementType type)
{
    newElements.insert(DrawElement {name, type});
}

const DrawColor DebugDrawerHelper::Colors::Red    = DrawColor {1, 0, 0, 1};
const DrawColor DebugDrawerHelper::Colors::Green  = DrawColor {0, 1, 0, 1};
const DrawColor DebugDrawerHelper::Colors::Blue   = DrawColor {0, 0, 1, 1};
const DrawColor DebugDrawerHelper::Colors::Yellow = DrawColor {1, 1, 0, 1};
const DrawColor DebugDrawerHelper::Colors::Orange = DrawColor {1, 0.5f, 0, 1};

DrawColor DebugDrawerHelper::Colors::Lerp(const DrawColor& a, const DrawColor& b, float f)
{
    return DrawColor
    {
        math::Helpers::Lerp(a.r, b.r, f),
        math::Helpers::Lerp(a.g, b.g, f),
        math::Helpers::Lerp(a.b, b.b, f),
        math::Helpers::Lerp(a.a, b.a, f)
    };
}
