/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2017, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarX
 * @author     Mirko Waechter( mirko.waechter at kit dot edu)
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "NJointCartesianActiveImpedanceController.h"
#include <VirtualRobot/RobotNodeSet.h>
#include "../RobotUnit.h"

#define DEFAULT_TCP_STRING "default TCP"

namespace armarx
{

    NJointControllerRegistration<NJointCartesianActiveImpedanceController> registrationControllerNJointCartesianActiveImpedanceController("NJointCartesianActiveImpedanceController");

    std::string NJointCartesianActiveImpedanceController::getClassName(const Ice::Current&) const
    {
        return "NJointCartesianActiveImpedanceController";
    }

    NJointCartesianActiveImpedanceController::NJointCartesianActiveImpedanceController(NJointControllerDescriptionProviderInterfacePtr prov, NJointCartesianActiveImpedanceControllerConfigPtr config, const boost::shared_ptr<VirtualRobot::Robot>& r)
    {
        ARMARX_CHECK_EXPRESSION(prov);
        RobotUnitPtr robotUnit = RobotUnitPtr::dynamicCast(prov);
        ARMARX_CHECK_EXPRESSION(robotUnit);
        ARMARX_CHECK_EXPRESSION(!config->nodeSetName.empty());
        VirtualRobot::RobotNodeSetPtr rns = robotUnit->getRtRobot()->getRobotNodeSet(config->nodeSetName);
        ARMARX_CHECK_EXPRESSION_W_HINT(rns, config->nodeSetName);
        for (size_t i = 0; i < rns->getSize(); ++i)
        {
            std::string jointName = rns->getNode(i)->getName();
            ControlTargetBase* ct = prov->getControlTarget(jointName, ControlModes::Velocity1DoF);
            const SensorValueBase* sv = prov->getSensorValue(jointName);
            targets.push_back(ct->asA<ControlTarget1DoFActuatorVelocity>());

            const SensorValue1DoFActuatorTorque* torqueSensor = sv->asA<SensorValue1DoFActuatorTorque>();
            const SensorValue1DoFGravityTorque* gravityTorqueSensor = sv->asA<SensorValue1DoFGravityTorque>();
            if (!torqueSensor)
            {
                ARMARX_WARNING << "No Torque sensor available for " << jointName;
            }
            if (!gravityTorqueSensor)
            {
                ARMARX_WARNING << "No Gravity Torque sensor available for " << jointName;
            }
            torqueSensors.push_back(torqueSensor);
            gravityTorqueSensors.push_back(gravityTorqueSensor);
        };

        tcp = (config->tcpName.empty() || config->tcpName == DEFAULT_TCP_STRING) ? rns->getTCP() : robotUnit->getRtRobot()->getRobotNode(config->tcpName);
        ARMARX_CHECK_EXPRESSION_W_HINT(tcp, config->tcpName);
        tcpController.reset(new CartesianPositionController(tcp));
        ik.reset(new VirtualRobot::DifferentialIK(rns, rns->getRobot()->getRootNode(), VirtualRobot::JacobiProvider::eSVDDamped));


        nodeSetName = config->nodeSetName;
    }


    void NJointCartesianActiveImpedanceController::rtPreActivateController()
    {
        targetPose = tcp->getPoseInRootFrame();
    }

    void NJointCartesianActiveImpedanceController::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {

        Eigen::VectorXf cartesianFT = tcpController->calculate(targetPose, VirtualRobot::IKSolver::CartesianSelection::All);
        Eigen::MatrixXf jacobi = ik->getJacobianMatrix(tcp, VirtualRobot::IKSolver::CartesianSelection::All);
        Eigen::MatrixXf jacobiT = jacobi.transpose();


        Eigen::VectorXf jointTargetTorques = jacobiT * cartesianFT;

        ARMARX_IMPORTANT << deactivateSpam(0.5) << VAROUT(jointTargetTorques.transpose());

        for (size_t i = 0; i < targets.size(); ++i)
        {
            targets.at(i)->torque = jointTargetTorques(i);
        }
    }


    ::armarx::WidgetDescription::WidgetSeq NJointCartesianActiveImpedanceController::createSliders()
    {
        using namespace armarx::WidgetDescription;
        ::armarx::WidgetDescription::WidgetSeq widgets;
        auto addSlider = [&](const std::string & label, float limit)
        {
            widgets.emplace_back(new Label(false, label));
            {
                FloatSliderPtr c_x = new FloatSlider;
                c_x->name = label;
                c_x->min = -limit;
                c_x->defaultValue = 0.0f;
                c_x->max = limit;
                widgets.emplace_back(c_x);
            }
        };

        addSlider("x", 100);
        addSlider("y", 100);
        addSlider("z", 100);
        addSlider("roll", 0.5);
        addSlider("pitch", 0.5);
        addSlider("yaw", 0.5);
        addSlider("avoidJointLimitsKp", 1);
        return widgets;
    }

    WidgetDescription::HBoxLayoutPtr NJointCartesianActiveImpedanceController::createJointSlidersLayout(float min, float max, float defaultValue) const
    {
        using namespace armarx::WidgetDescription;
        HBoxLayoutPtr layout = new HBoxLayout;
        auto addSlider = [&](const std::string & label)
        {
            layout->children.emplace_back(new Label(false, label));
            FloatSliderPtr floatSlider = new FloatSlider;
            floatSlider->name = label;
            floatSlider->min = min;
            floatSlider->defaultValue = defaultValue;
            floatSlider->max = max;
            layout->children.emplace_back(floatSlider);
        };

        for (const VirtualRobot::RobotNodePtr& rn : tcpController->rns->getAllRobotNodes())
        {
            addSlider(rn->getName());
        }

        return layout;

    }

    WidgetDescription::WidgetPtr NJointCartesianActiveImpedanceController::GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>& robot, const std::map<std::string, ConstControlDevicePtr>& controlDevices, const std::map<std::string, ConstSensorDevicePtr>&)
    {
        using namespace armarx::WidgetDescription;
        HBoxLayoutPtr layout = new HBoxLayout;

        LabelPtr label = new Label;
        label->text = "nodeset name";
        layout->children.emplace_back(label);
        StringComboBoxPtr box = new StringComboBox;
        box->defaultIndex = 0;

        box->options = robot->getRobotNodeSetNames();
        box->name = "nodeSetName";
        layout->children.emplace_back(box);

        LabelPtr labelTCP = new Label;
        labelTCP->text = "tcp name";
        layout->children.emplace_back(labelTCP);
        StringComboBoxPtr boxTCP = new StringComboBox;
        boxTCP->defaultIndex = 0;

        boxTCP->options = robot->getRobotNodeNames();
        boxTCP->options.insert(boxTCP->options.begin(), DEFAULT_TCP_STRING);
        boxTCP->name = "tcpName";
        layout->children.emplace_back(boxTCP);

        layout->children.emplace_back(new HSpacer);
        return layout;
    }

    NJointCartesianActiveImpedanceControllerConfigPtr NJointCartesianActiveImpedanceController::GenerateConfigFromVariants(const StringVariantBaseMap& values)
    {
        return new NJointCartesianActiveImpedanceControllerConfig(values.at("nodeSetName")->getString(), values.at("tcpName")->getString());
    }



    std::string NJointCartesianActiveImpedanceController::getNodeSetName() const
    {
        return nodeSetName;
    }

    void NJointCartesianActiveImpedanceController::rtPostDeactivateController()
    {

    }

    WidgetDescription::StringWidgetDictionary NJointCartesianActiveImpedanceController::getFunctionDescriptions(const Ice::Current&) const
    {
        using namespace armarx::WidgetDescription;
        HBoxLayoutPtr layout = new HBoxLayout;
        auto sliders = createSliders();
        layout->children.insert(layout->children.end(),
                                sliders.begin(),
                                sliders.end());

        return
        {
            {"ControllerTarget", layout}
        };
    }

    void NJointCartesianActiveImpedanceController::callDescribedFunction(const std::string& name, const StringVariantBaseMap& valueMap, const Ice::Current&)
    {
        if (name == "ControllerTarget")
        {
            /*LockGuardType guard {controlDataMutex};
            getWriterControlStruct().xVel = valueMap.at("x")->getFloat();
            getWriterControlStruct().yVel = valueMap.at("y")->getFloat();
            getWriterControlStruct().zVel = valueMap.at("z")->getFloat();
            getWriterControlStruct().rollVel = valueMap.at("roll")->getFloat();
            getWriterControlStruct().pitchVel = valueMap.at("pitch")->getFloat();
            getWriterControlStruct().yawVel = valueMap.at("yaw")->getFloat();
            getWriterControlStruct().avoidJointLimitsKp = valueMap.at("avoidJointLimitsKp")->getFloat();
            writeControlStruct();*/
        }
        else
        {
            ARMARX_WARNING << "Unknown function name called: " << name;
        }
    }

    float SimplePID::update(float dt, float error)
    {
        float derivative = (error - lastError) / dt;
        float retVal = Kp * error + Kd * derivative;
        lastError = error;
        return retVal;
    }

} // namespace armarx



void armarx::NJointCartesianActiveImpedanceController::setControllerTarget(float x, float y, float z, float qw, float qx, float qy, float qz, const Ice::Current&)
{
    LockGuardType guard {controlDataMutex};
    getWriterControlStruct().xPos = x;
    getWriterControlStruct().yPos = y;
    getWriterControlStruct().zVepos = z;
    getWriterControlStruct().qw = qw;
    getWriterControlStruct().qx = qx;
    getWriterControlStruct().qy = qy;
    getWriterControlStruct().qz = qz;
    writeControlStruct();
}

