/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "DataFieldsInfo.h"

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <ArmarXCore/observers/variant/TimestampVariant.h>
#include <RobotAPI/libraries/core/Pose.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/core/OrientedPoint.h>

#include <RobotAPI/interface/units/KinematicUnitInterface.h>


namespace armarx
{
    namespace introspection
    {
        std::string DataFieldsInfo<Eigen::Vector3f, void>::GetFieldAsString(std::size_t i, const Eigen::Vector3f& field)
        {
            ARMARX_CHECK_LESS(i, 3);
            return to_string(field(i));
        }

        std::map<std::string, VariantBasePtr> DataFieldsInfo<Eigen::Vector3f, void>::ToVariants(
            const Eigen::Vector3f& value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            if (!frame.empty())
            {
                return {{name, {new TimedVariant(FramedPosition{value, frame, agent}, timestamp)}}};
            }
            ARMARX_CHECK_EXPRESSION_W_HINT(agent.empty(), "No frame but an agent given");
            return {{name, {new TimedVariant(Vector3{value}, timestamp)}}};
        }
    }
}

namespace armarx
{
    namespace introspection
    {
#define make_DataFieldsInfo_for_eigen_vector(Type,Num)                                                                              \
    std::string DataFieldsInfo<Eigen::Vector##Num##Type, void>::GetFieldAsString(std::size_t i, const Eigen::Vector##Num##Type& field)  \
    {                                                                                                                                   \
        ARMARX_CHECK_LESS(i, Num);                                                                                                      \
        return to_string(field(i));                                                                                                \
    }                                                                                                                                   \
    std::map<std::string, VariantBasePtr> DataFieldsInfo<Eigen::Vector##Num##Type, void>::ToVariants(                                   \
            const Eigen::Vector##Num##Type& value,                                                                                          \
            const std::string& name,                                                                                                        \
            const IceUtil::Time& timestamp,                                                                                                 \
            const std::string& frame,                                                                                                       \
            const std::string& agent)                                                                                                       \
    {                                                                                                                                   \
        ARMARX_CHECK_EXPRESSION_W_HINT(frame.empty() && agent.empty(), "There is no framed version of TimestampVariant");               \
        std::map<std::string, VariantBasePtr> result;                                                                                   \
        for(int i = 0; i < Num; ++i)                                                                                                    \
        {                                                                                                                               \
            result.emplace(name + "." + to_string(i), VariantBasePtr{new TimedVariant(value(i), timestamp)});                      \
        }                                                                                                                               \
        return result;                                                                                                                  \
    }                                                                                                                                   \
    std::vector<std::string> DataFieldsInfo<Eigen::Vector##Num##Type, void>::GetFieldNames()                                            \
    {                                                                                                                                   \
        std::vector<std::string> result;                                                                                                \
        for(int i = 0; i < Num; ++i)                                                                                                    \
        {                                                                                                                               \
            result.emplace_back(to_string(i));                                                                                     \
        }                                                                                                                               \
        return result;                                                                                                                  \
    }

        make_DataFieldsInfo_for_eigen_vector(f, 2)
        make_DataFieldsInfo_for_eigen_vector(f, 4)
        make_DataFieldsInfo_for_eigen_vector(f, 5)
        make_DataFieldsInfo_for_eigen_vector(f, 6)

        make_DataFieldsInfo_for_eigen_vector(d, 2)
        make_DataFieldsInfo_for_eigen_vector(d, 3)
        make_DataFieldsInfo_for_eigen_vector(d, 4)
        make_DataFieldsInfo_for_eigen_vector(d, 5)
        make_DataFieldsInfo_for_eigen_vector(d, 6)

        make_DataFieldsInfo_for_eigen_vector(i, 2)
        make_DataFieldsInfo_for_eigen_vector(i, 3)
        make_DataFieldsInfo_for_eigen_vector(i, 4)
        make_DataFieldsInfo_for_eigen_vector(i, 5)
        make_DataFieldsInfo_for_eigen_vector(i, 6)
#undef make_DataFieldsInfo_for_eigen_vector
    }
}

namespace armarx
{
    namespace introspection
    {
        std::string DataFieldsInfo<JointStatus, void>::GetFieldAsString(
            std::size_t i, const JointStatus& field)
        {
            ARMARX_CHECK_LESS(i, 4);
            switch (i)
            {
                case 0:
                    return to_string(field.error);
                case 1:
                    return to_string(field.operation);
                case 2:
                    return to_string(field.enabled);
                case 3:
                    return to_string(field.emergencyStop);
            }
            throw std::logic_error
            {
                __FILE__ " : " + to_string(__LINE__) +
                " : Unreachable code reached"
            };
        }

        std::map<std::string, VariantBasePtr> DataFieldsInfo<JointStatus, void>::ToVariants(
            const JointStatus& value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            ARMARX_CHECK_EXPRESSION_W_HINT(frame.empty() && agent.empty(), "There is no framed version for JointStatus");
            static const auto fnames = GetFieldNames();
            return
            {
                {name + fnames.at(0), {new TimedVariant{to_string(value.error), timestamp}}},
                {name + fnames.at(1), {new TimedVariant{to_string(value.operation), timestamp}}},
                {name + fnames.at(2), {new TimedVariant{value.enabled, timestamp}}},
                {name + fnames.at(3), {new TimedVariant{value.emergencyStop, timestamp}}}
            };
        }
    }
}

namespace armarx
{
    namespace introspection
    {
        std::string DataFieldsInfo<Eigen::Quaternionf, void>::GetFieldAsString(
            std::size_t i,
            const Eigen::Quaternionf& field)
        {
            ARMARX_CHECK_LESS(i, 4);
            switch (i)
            {
                case 0:
                    return to_string(field.w());
                case 1:
                    return to_string(field.x());
                case 2:
                    return to_string(field.y());
                case 3:
                    return to_string(field.z());
            }
            throw std::logic_error
            {
                __FILE__ " : " + to_string(__LINE__) +
                " : Unreachable code reached"
            };
        }

        std::map<std::string, VariantBasePtr> DataFieldsInfo<Eigen::Quaternionf, void>::ToVariants(
            const Eigen::Quaternionf& value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            if (!frame.empty())
            {
                return {{name, {new TimedVariant(FramedOrientation{value, frame, agent}, timestamp)}}};
            }
            ARMARX_CHECK_EXPRESSION_W_HINT(agent.empty(), "No frame but an agent given");
            return {{name, {new TimedVariant(Quaternion{value}, timestamp)}}};
        }
    }
}

namespace armarx
{
    namespace introspection
    {
        std::string DataFieldsInfo<Eigen::Matrix4f, void>::GetFieldAsString(std::size_t i, const Eigen::Matrix4f& field)
        {
            ARMARX_CHECK_LESS(i, 16);
            return  to_string(field(i / 4, i % 4));
        }

        std::map<std::string, VariantBasePtr> DataFieldsInfo<Eigen::Matrix4f, void>::ToVariants(
            const Eigen::Matrix4f& value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            if (!frame.empty())
            {
                return {{name, {new TimedVariant(FramedPose{value, frame, agent}, timestamp)}}};
            }
            ARMARX_CHECK_EXPRESSION_W_HINT(agent.empty(), "No frame but an agent given");
            return {{name, {new TimedVariant(Pose{value}, timestamp)}}};
        }
    }
}

namespace armarx
{
    namespace introspection
    {
        std::map<std::string, VariantBasePtr> DataFieldsInfo<std::chrono::microseconds, void>::ToVariants(
            std::chrono::microseconds value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            ARMARX_CHECK_EXPRESSION_W_HINT(frame.empty() && agent.empty(), "There is no framed version of TimestampVariant");
            return {{name, {new TimedVariant(TimestampVariant{value.count()}, timestamp)}}};
        }

        std::map<std::string, VariantBasePtr> DataFieldsInfo<IceUtil::Time, void>::ToVariants(
            IceUtil::Time value,
            const std::string& name,
            const IceUtil::Time& timestamp,
            const std::string& frame,
            const std::string& agent)
        {
            ARMARX_CHECK_EXPRESSION_W_HINT(frame.empty() && agent.empty(), "There is no framed version of TimestampVariant");
            return {{name, {new TimedVariant(TimestampVariant{value.toMicroSeconds()}, timestamp)}}};
        }
    }
}
