/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::CartesianWaypointControlGuiWidgetController
 * \author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * \date       2019
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <random>
#include <string>

#include <VirtualRobot/Util/json.h>

#include <ArmarXCore/util/CPPUtility/container.h>
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>

#include "CartesianWaypointControlGuiWidgetController.h"

namespace armarx
{
    CartesianWaypointControlGuiWidgetController::CartesianWaypointControlGuiWidgetController()
    {
        _ui.setupUi(getWidget());
        connect(_ui.radioButtonWPJson, SIGNAL(clicked()), this, SLOT(triggerParsing()));
        connect(_ui.radioButton4f, SIGNAL(clicked()), this, SLOT(triggerParsing()));
        connect(_ui.textEditWPs, SIGNAL(textChanged()), this, SLOT(triggerParsing()));

        connect(_ui.pushButtonStop,             SIGNAL(clicked()), this, SLOT(on_pushButtonStop_clicked()));
        connect(_ui.pushButtonExecute,          SIGNAL(clicked()), this, SLOT(on_pushButtonExecute_clicked()));
        connect(_ui.pushButtonZeroFT,           SIGNAL(clicked()), this, SLOT(on_pushButtonZeroFT_clicked()));
        connect(_ui.pushButtonSendSettings,     SIGNAL(clicked()), this, SLOT(on_pushButtonSendSettings_clicked()));
        connect(_ui.pushButtonCreateController, SIGNAL(clicked()), this, SLOT(on_pushButtonCreateController_clicked()));
    }

    CartesianWaypointControlGuiWidgetController::~CartesianWaypointControlGuiWidgetController() = default;


    void CartesianWaypointControlGuiWidgetController::loadSettings(QSettings* settings)
    {
        _robotStateComponentName = settings->value("rsc", "Armar6StateComponent").toString().toStdString();
        _robotUnitName = settings->value("ru", "Armar6Unit").toString().toStdString();
    }

    void CartesianWaypointControlGuiWidgetController::saveSettings(QSettings* settings)
    {
        settings->setValue("rsc", QString::fromStdString(_robotStateComponentName));
        settings->setValue("ru", QString::fromStdString(_robotUnitName));
    }


    void CartesianWaypointControlGuiWidgetController::onInitComponent()
    {
        usingProxy(_robotStateComponentName);
        usingProxy(_robotUnitName);
    }


    void CartesianWaypointControlGuiWidgetController::onConnectComponent()
    {
        //proxies
        {
            _robotStateComponent = getProxy<RobotStateComponentInterfacePrx>(_robotStateComponentName);
            _robotUnit = getProxy<RobotUnitInterfacePrx>(_robotUnitName);
        }
        //robot
        {
            _robot = RemoteRobot::createLocalCloneFromFile(_robotStateComponent, VirtualRobot::RobotIO::eStructure);
        }
        //fill rns combo box
        {
            bool found = false;
            for (const auto& rnsn : _robot->getRobotNodeSetNames())
            {
                _ui.comboBoxChain->addItem(QString::fromStdString(rnsn));
                if (rnsn == "RightArm")
                {
                    _ui.comboBoxChain->setCurrentIndex(_ui.comboBoxChain->count() - 1);
                    found = true;
                }
            }
            if (found && _robot->hasRobotNode("FT R"))
            {
                _ui.lineEditFTName->setText("FT R");
            }
        }
    }
    void CartesianWaypointControlGuiWidgetController::onDisconnectComponent()
    {
        deleteOldController();
    }
    void CartesianWaypointControlGuiWidgetController::on_pushButtonCreateController_clicked()
    {
        deleteOldController();

        static const std::string njointControllerClassName = "NJointCartesianWaypointController";

        NJointCartesianWaypointControllerConfigPtr cfg = new NJointCartesianWaypointControllerConfig;
        cfg->rns = _ui.comboBoxChain->currentText().toStdString();
        cfg->ftName = _ui.lineEditFTName->text().toStdString();
        cfg->runCfg = readRunCfg();
        _supportsFT = !cfg->ftName.empty();

        _controllerName = getName() + "_" + njointControllerClassName +
                          std::to_string(std::random_device{}());

        _ui.labelCurrentControllerChain->setText(QString::fromStdString(cfg->rns));
        _ui.labelCurrentControllerFT->setText(QString::fromStdString(cfg->ftName));

        ARMARX_IMPORTANT << "Creating " << njointControllerClassName << " '"
                         << _controllerName << "'";
        _controller = NJointCartesianWaypointControllerInterfacePrx::checkedCast(
                          _robotUnit->createNJointController(
                              njointControllerClassName,
                              _controllerName,
                              cfg));
        ARMARX_CHECK_NOT_NULL(_controller);
    }

    void CartesianWaypointControlGuiWidgetController::on_pushButtonSendSettings_clicked()
    {
        if (_controller)
        {
            ARMARX_IMPORTANT << "sending new config to " << _controllerName;
            _controller->setConfig(readRunCfg());
        }
    }

    void CartesianWaypointControlGuiWidgetController::on_pushButtonZeroFT_clicked()
    {
        if (_controller && _supportsFT)
        {
            ARMARX_IMPORTANT << "setting ft offset for " << _controllerName;
            _controller->setCurrentFTAsOffset();
        }
    }

    void CartesianWaypointControlGuiWidgetController::on_pushButtonExecute_clicked()
    {
        if (_controller)
        {
            _controller->activateController();
            ARMARX_IMPORTANT << "trigger execution of " << _lastParsedWPs.size()
                             << " waypoints on " << _controllerName;
            _controller->setWaypoints(_ui.checkBoxWPReverse->isChecked() ? reverse(_lastParsedWPs) : _lastParsedWPs);
        }
    }

    void CartesianWaypointControlGuiWidgetController::on_pushButtonStop_clicked()
    {
        if (_controller)
        {
            ARMARX_IMPORTANT << "sending stop command to controller " << _controllerName;
            _controller->stopMovement();
        }
    }

    QPointer<QDialog> CartesianWaypointControlGuiWidgetController::getConfigDialog(QWidget* parent)
    {
        if (!_dialog)
        {
            _dialog = new SimpleConfigDialog(parent);
            _dialog->addProxyFinder<RobotUnitInterfacePrx>({"ru", "Robot Unit", "*Unit"});
            _dialog->addProxyFinder<RobotStateComponentInterfacePrx>({"rsc", "Robot State Component", "*Component"});
        }
        return qobject_cast<SimpleConfigDialog*>(_dialog);
    }

    void CartesianWaypointControlGuiWidgetController::configured()
    {
        _robotStateComponentName = _dialog->getProxyName("rsc");
        _robotUnitName = _dialog->getProxyName("ru");
    }

    void CartesianWaypointControlGuiWidgetController::triggerParsing()
    {
        _lastParsedWPs.clear();
        _ui.labelParsingSuccess->setText("<pre parsing>");
        if (_ui.radioButtonWPJson->isChecked())
        {
            //parse json
            try
            {
                _lastParsedWPs = VirtualRobot::json::posquatArray2eigen4fVector(
                                     _ui.textEditWPs->toPlainText().toStdString());
            }
            catch (...)
            {
                _ui.labelParsingSuccess->setText("Failed to parse json array!");
                return;
            }
        }
        else if (_ui.radioButton4f->isChecked())
        {
            //parse lines
            ///TODO parse Matrix4f style input
            _ui.labelParsingSuccess->setText("NYI");
            return;
        }
        //test reachability
        {
            ///TODO test reachability and visualize
        }
        _ui.labelParsingSuccess->setText("Parsed " + QString::number(_lastParsedWPs.size()) + " waypoints.");
    }

    NJointCartesianWaypointControllerRuntimeConfig CartesianWaypointControlGuiWidgetController::readRunCfg() const
    {
        NJointCartesianWaypointControllerRuntimeConfig cfg;

        cfg.wpCfg.maxPositionAcceleration     = static_cast<float>(_ui.doubleSpinBoxMaxAccPos->value());
        cfg.wpCfg.maxOrientationAcceleration  = static_cast<float>(_ui.doubleSpinBoxMaxAccOri->value());
        cfg.wpCfg.maxNullspaceAcceleration    = static_cast<float>(_ui.doubleSpinBoxMaxAccNull->value());

        cfg.wpCfg.kpJointLimitAvoidance       = static_cast<float>(_ui.doubleSpinBoxLimitAvoidKP->value());
        cfg.wpCfg.jointLimitAvoidanceScale    = static_cast<float>(_ui.doubleSpinBoxLimitAvoidScale->value());

        cfg.wpCfg.thresholdOrientationNear    = static_cast<float>(_ui.doubleSpinBoxOriNear->value());
        cfg.wpCfg.thresholdOrientationReached = static_cast<float>(_ui.doubleSpinBoxOriReached->value());
        cfg.wpCfg.thresholdPositionNear       = static_cast<float>(_ui.doubleSpinBoxPosNear->value());
        cfg.wpCfg.thresholdPositionReached    = static_cast<float>(_ui.doubleSpinBoxPosReached->value());

        cfg.wpCfg.maxOriVel                   = static_cast<float>(_ui.doubleSpinBoxOriMaxVel->value());
        cfg.wpCfg.maxPosVel                   = static_cast<float>(_ui.doubleSpinBoxPosMaxVel->value());
        cfg.wpCfg.kpOri                       = static_cast<float>(_ui.doubleSpinBoxOriKP->value());
        cfg.wpCfg.kpPos                       = static_cast<float>(_ui.doubleSpinBoxPosKP->value());

        cfg.forceThreshold                    = static_cast<float>(_ui.doubleSpinBoxFTLimit->value());
        cfg.optimizeNullspaceIfTargetWasReached = _ui.checkBoxKeepOptimizing->isChecked();

        return cfg;
    }

    void CartesianWaypointControlGuiWidgetController::deleteOldController()
    {
        if (_controller)
        {
            ARMARX_IMPORTANT << "deleting old controller '" << _controllerName << "'";
            try
            {
                _controller->deactivateAndDeleteController();
            }
            catch (...)
            {
                _controller = nullptr;
            }
        }
    }
}


