#include <chrono>

#include <ArmarXCore/core/logging/Logging.h>
#include "JointPWMVelocityControllerV2.h"
#include "../KITGripperBasisBoard.h"
#include <RobotAPI/components/units/RobotUnit/util/ControlThreadOutputBuffer.h>
#include <ArmarXCore/core/ManagedIceObject.h>


#include <ArmarXGui/libraries/RemoteGui/WidgetProxy.h>
#include <ArmarXGui/libraries/RemoteGui/WidgetBuilder.h>

#include <ArmarXCore/core/application/Application.h>

#include <RobotAPI/components/units/RobotUnit/RobotUnitModules/RobotUnitModuleDevices.h>

using namespace armarx;


JointPWMVelocityControllerV2::JointPWMVelocityControllerV2(const std::string deviceName, KITGripperBasisBoardPtr board, ActorDataV2Ptr jointData,
        PWMVelocityControllerConfigurationCPtr velocityControllerConfigDataPtr) : JointController(),
    config(velocityControllerConfigDataPtr),
    controller(velocityControllerConfigDataPtr),
    target(), board(board), deviceName(deviceName)
{
    actorIndex = board->getActorIndex(deviceName);
    dataPtr = jointData;

    velController.acceleration = velocityControllerConfigDataPtr->maxAccelerationRad;
    velController.deceleration = velocityControllerConfigDataPtr->maxDecelerationRad;
    velController.maxDt = velocityControllerConfigDataPtr->maxDt;
    velController.maxV = velocityControllerConfigDataPtr->maxVelocityRad;
    velController.directSetVLimit = velocityControllerConfigDataPtr->directSetVLimit;
    ARMARX_CHECK_GREATER_EQUAL(jointData->getSoftLimitHi(), jointData->getSoftLimitLo());
    //    controller.positionLimitHiHard = dataPtr->getHardLimitHi();
    velController.positionLimitHiSoft = jointData->getSoftLimitHi();
    //    controller.positionLimitLoHard = dataPtr->getHardLimitLo();
    velController.positionLimitLoSoft = jointData->getSoftLimitLo();
    this->isLimitless = jointData->isLimitless();
}

JointPWMVelocityControllerV2::~JointPWMVelocityControllerV2() noexcept(true)
{
    stopRequested = true;
    try
    {
        threadHandle.join();
    }
    catch (...)
    {

    }
}

void JointPWMVelocityControllerV2::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
{
    if (target.isValid())
    {

        {
            auto currentPosition = dataPtr->getPosition();
            if (isLimitless)
            {
                velController.currentPosition = velController.positionLimitHiSoft - (velController.positionLimitHiSoft - velController.positionLimitLoSoft) * 0.5;
                //                ARMARX_INFO << VAROUT(velController.currentPosition) << VAROUT(velController.positionLimitLoSoft) << VAROUT(velController.positionLimitHiSoft);
            }
            else
            {
                velController.currentPosition =  currentPosition;
            }
            velController.currentV = lastTargetVelocity;
            velController.dt = timeSinceLastIteration.toSecondsDouble();
            velController.targetV = target.velocity;
            double newVel = velController.run();


            //            ARMARX_INFO << deactivateSpam(1) << VAROUT(newVel) << VAROUT(target.velocity);
            if (std::isnan(newVel))
            {
                newVel = 0;
            }
            //            float newVel = target.velocity;
            if ((currentPosition > velController.positionLimitHiSoft && target.velocity > 0)
                || (currentPosition < velController.positionLimitLoSoft && target.velocity < 0))
            {
                newVel = 0;
                ARMARX_INFO << deactivateSpam(1) << "Breaking now at " << dataPtr->getPosition() << " pwm: " << dataPtr->getTargetPWM();
            }


            auto targetPWM = static_cast<int>(controller.run(timeSinceLastIteration, dataPtr->getVelocity(), newVel));

            dataPtr->setTargetPWM(targetPWM);

            lastTargetVelocity = newVel;

            //            ARMARX_RT_LOGF_INFO("target velocity: %.3f, current velocity: %.3f, target pwm: %d, kp: %.3f ki: %f, kd: %f, max acc: %.3f",
            //                                target.velocity, dataPtr->getVelocity(), targetPWM, pid->Kp, pid->Ki, pid->Kd, controller.acceleration).deactivateSpam(1);

        }
    }
    else
    {
        ARMARX_ERROR << "invalid target set for actor";
    }
}

ControlTargetBase* JointPWMVelocityControllerV2::getControlTarget()
{
    return &target;
}

void JointPWMVelocityControllerV2::rtPreActivateController()
{
    lastTargetVelocity = dataPtr->getVelocity();
    controller.reset(dataPtr->getVelocity());
}

void JointPWMVelocityControllerV2::rtPostDeactivateController()
{
    ARMARX_RT_LOGF_INFO("Setting PWM to 0");
    dataPtr->setTargetPWM(0);
}

StringVariantBaseMap JointPWMVelocityControllerV2::publish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer) const
{

    if (!remoteGui && !threadHandle.isValid())
    {
        //        threadHandle = Application::getInstance()->getThreadPool()->runTask([this]
        //        {
        //            std::string guiTabName;
        //            while (!stopRequested)
        //            {
        //                ManagedIceObjectPtr object;
        //                ARMARX_IMPORTANT << deactivateSpam(1) << "Trying to get parent";
        //                try
        //                {
        //                    object = ManagedIceObjectPtr::dynamicCast(getParent().getOwner());
        //                    ARMARX_CHECK_EXPRESSION(object);
        //                    remoteGui = object->getProxy<RemoteGuiInterfacePrx>("RemoteGuiProvider", false, "", false);
        //                    if (!remoteGui)
        //                    {
        //                        return;
        //                    }
        //                    ARMARX_IMPORTANT << deactivateSpam(1) << "Got Proxy";
        //                    guiTabName = getParent().getDeviceName() + getControlMode();
        //                    break;
        //                }
        //                catch (...)
        //                {
        //                    handleExceptions();
        //                    sleep(1);
        //                }

        //            }
        //            if (remoteGui)
        //            {
        //                ARMARX_IMPORTANT << "Creating GUI " << guiTabName;
        //                using namespace RemoteGui;



        //                auto vLayout = makeVBoxLayout();

        //                {
        //                    WidgetPtr KpLabel = makeTextLabel("Kp: ");

        //                    WidgetPtr KiSlider = makeFloatSlider("KpSlider")
        //                                         .min(0.0f).max(5000.0f)
        //                                         .value(config->p);
        //                    WidgetPtr line = makeHBoxLayout()
        //                                     .children({KpLabel, KiSlider});

        //                    vLayout.addChild(line);

        //                }


        //                {
        //                    WidgetPtr KiLabel = makeTextLabel("Ki: ");
        //                    WidgetPtr KiSlider = makeFloatSlider("KiSlider")
        //                                         .min(0.0f).max(50000.0f)
        //                                         .value(config->i);

        //                    WidgetPtr line = makeHBoxLayout()
        //                                     .children({KiLabel, KiSlider});

        //                    vLayout.addChild(line);

        //                }

        //                {
        //                    WidgetPtr KdLabel = makeTextLabel("Kd: ");
        //                    WidgetPtr KdSlider = makeFloatSlider("KdSlider")
        //                                         .min(0.0f).max(50.0f)
        //                                         .steps(100)
        //                                         .value(config->d);

        //                    WidgetPtr line = makeHBoxLayout()
        //                                     .children({KdLabel, KdSlider});

        //                    vLayout.addChild(line);
        //                    vLayout.addChild(new VSpacer);
        //                }

        //                //        WidgetPtr spin = makeFloatSpinBox("KpSpin")
        //                //                         .min(0.0f).max(2.0f)
        //                //                         .steps(20).decimals(2)
        //                //                         .value(0.4f);




        //                WidgetPtr groupBox = makeGroupBox("GroupBox")
        //                                     .label("Group")
        //                                     .child(vLayout);

        //                remoteGui->createTab(guiTabName, groupBox);

        //                while (!stopRequested)
        //                {
        //                    RemoteGui::TabProxy tab(remoteGui, guiTabName);
        //                    tab.receiveUpdates();
        //                    this->controller.pid->Kp = tab.getValue<float>("KpSlider").get();
        //                    this->controller.pid->Ki = tab.getValue<float>("KiSlider").get();
        //                    this->controller.pid->Kd = tab.getValue<float>("KdSlider").get();
        //                    usleep(100000);
        //                }
        //            }

        //        });
    }
    return {{"lastTargetVelocity", new Variant(lastTargetVelocity.load())},
        {"filteredVelocity", new Variant(controller.lastActualVelocity.load())},
        {"pidIntegralCV", new Variant(controller.pid->integral * controller.pid->Ki)},
        {"pidPropCV", new Variant(controller.pid->previousError * controller.pid->Kp)},
        {"pidDiffCV", new Variant(controller.pid->derivative * controller.pid->Kd)}

    };
}





