/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2011-2017, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarX
 * @author     Mirko Waechter( mirko.waechter at kit dot edu)
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#pragma once

#include <RobotAPI/libraries/ArmarXEtherCAT/AbstractData.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValueIMU.h>
#include <VirtualRobot/Nodes/RobotNode.h>
#include <ArmarXCore/observers/filters/rtfilters/AverageFilter.h>
#include "KITGripperBasisBoardSlave.h"

namespace armarx
{
    class   SensorValueKITGripperBasisBoard :
    //            virtual public SensorValue1DoFActuatorMotorTemperature,
        virtual public SensorValueIMU
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        int IMUTemperature;
        static SensorValueInfo<SensorValueKITGripperBasisBoard> GetClassMemberInfo()
        {
            SensorValueInfo<SensorValueKITGripperBasisBoard> svi;
            //            svi.addBaseClass<SensorValue1DoFActuatorMotorTemperature>();
            svi.addMemberVariable(&SensorValueKITGripperBasisBoard::IMUTemperature, "IMUTemperature");
            svi.addBaseClass<SensorValueIMU>();
            return svi;
        }
        SensorValueKITGripperBasisBoard() = default;
        SensorValueKITGripperBasisBoard(SensorValueKITGripperBasisBoard&&) = default;
        SensorValueKITGripperBasisBoard(const SensorValueKITGripperBasisBoard&) = default;
        SensorValueKITGripperBasisBoard& operator=(SensorValueKITGripperBasisBoard&& other)
        {
            *this = other;
            return *this;
        }
        SensorValueKITGripperBasisBoard& operator=(const SensorValueKITGripperBasisBoard&) = default;

    };

    class KITGripperActorSensorData : virtual public SensorValue1DoFRealActuator
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        int32_t targetPWM;
        float relativePosition;
        float velocityTicksPerMs;
        float absoluteEncoderVelocity;
        int32_t maxPWM;
        int32_t minPWM;
        float torqueWithInertia;
        float rawVelocity;
        float filteredVelocity;

        static SensorValueInfo<KITGripperActorSensorData> GetClassMemberInfo()
        {
            SensorValueInfo<KITGripperActorSensorData> svi;
            //            svi.addBaseClass<SensorValue1DoFActuatorMotorTemperature>();
            svi.addBaseClass<SensorValue1DoFRealActuator>();
            svi.addMemberVariable(&KITGripperActorSensorData::targetPWM, "targetPWM");
            svi.addMemberVariable(&KITGripperActorSensorData::relativePosition, "relativePosition");
            svi.addMemberVariable(&KITGripperActorSensorData::maxPWM, "maxPWM");
            svi.addMemberVariable(&KITGripperActorSensorData::minPWM, "minPWM");
            svi.addMemberVariable(&KITGripperActorSensorData::velocityTicksPerMs, "velocityTicksPerMs");
            svi.addMemberVariable(&KITGripperActorSensorData::absoluteEncoderVelocity, "absoluteEncoderVelocity");
            svi.addMemberVariable(&KITGripperActorSensorData::torqueWithInertia, "torqueWithInertia");
            svi.addMemberVariable(&KITGripperActorSensorData::rawVelocity, "rawVelocity");
            svi.addMemberVariable(&KITGripperActorSensorData::filteredVelocity, "filteredVelocity");
            return svi;
        }

    };


    class ActorData
    {
    public:
        ActorData();
        ActorData(int32_t filterWindowSize);
        float getPosition() const;
        float getRelativePosition() const;
        float getCompensatedRelativePosition() const;
        float getAcceleration() const;
        int32_t getRelativePositionSensorOnly() const;
        bool getPositionControlEnabled() const;
        float getSoftLimitHi() const;
        float getSoftLimitLo() const;
        bool isLimitless() const;

        // velocity data function
        float getVelocity() const;
        float getRawVelocity() const;
        int32_t getUseABSEncoderForVelocity() const;
        bool getVelocityControlEnabled() const;
        float getAbsoluteEncoderVelocity() const;
        int32_t getVelocityTicks() const;

        // torque data function
        float getTorque() const;
        bool getTorqueControlEnabled() const;

        // pwm data function
        void setTargetPWM(int32_t targetPWM);
        int32_t getTargetPWM() const;
        int32_t getCurrentMinPWM() const;
        int32_t getCurrentMaxPWM() const;

        float getCurrentPWMBoundGradient() const;
        int32_t getCurrentPWMBoundOffset() const;
        std::string getMotorModelTypeIndex() const;
        bool getNonlinearModelEnabled() const;
        float getMotorInertia() const;

        // parallel gripper data function
        int32_t getSiblingControlActorIndex() const;
        float getParallelGripperDecouplingFactor() const;



    private:
        // parallel gripper controller configuration
        float parallelGripperDecouplingFactor = std::nanf("");
        int32_t parallelControlSiblingIndex = -1; // meaning: -1, not coupled with other motors, (0 or 1 or 2): coupled with respective motors on the same board

        // position & relative position data
        bool positionControlEnabled = true;
        u_int32_t rawABSEncoderTicks;
        LinearConvertedValue<int32_t> relativePosition;
        float adjustedRelativePosition;
        float relativePositionOffset = std::nan("");
        LinearConvertedValue<u_int32_t> position;
        float sanitizedAbsolutePosition;
        float lastAbsolutePosition = std::nanf("");
        int32_t relativePositionSensorOnly = -1;

        // velocity data
        bool velocityControlEnabled = true;
        LinearConvertedValue<int32_t> velocity;
        float rawVelocity = 0.0f;
        //        int32_t velocityFilterWindowSize = 10;
        rtfilters::AverageFilter velocityFilter;
        float absoluteEncoderVelocity = 0.0f;
        float acceleration;
        int32_t* velocityTicks;
        int32_t useABSEncoderForVelocity = -1;

        // torque data
        //        LinearConvertedValue<int32_t> torque;
        bool torqueControlEnabled = true;
        std::string motorModelType = "None";
        bool nonlinearModelEnabled = 0;
        float motorInertia = 0.0;

        // pwm data
        LinearConvertedValue<int32_t> targetPWM;
        int32_t currentMaxPWM = 0;
        int32_t currentMinPWM = 0;
        size_t maxPWM;
        LinearConvertedValue<int32_t> targetPWMPtr;
        float currentPWMBoundGradient = 3.75;
        int32_t currentPWMBoundOffset = 1500;

        VirtualRobot::RobotNodePtr robotNode;

        friend class KITGripperBasisBoardData;
    };
    using ActorDataPtr = std::shared_ptr<ActorData>;

    class KITGripperBasisBoardData : public AbstractData
    {
    public:
        KITGripperBasisBoardData(const RapidXmlReaderNode& node, DefaultRapidXmlReaderNode defaultConfigurationNode,
                                 KITGripperBasisBoardOUT_t* sensorOUT, KITGripperBasisBoardIN_t* sensorIN, VirtualRobot::RobotPtr robot);

        // AbstractData interface
    public:
        void rtReadSensorValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;
        void rtWriteTargetValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;
        ActorDataPtr& getActorData(size_t actorIndex);

        int8_t getIMUTemperature() const;

    private:
        KITGripperBasisBoardOUT_t* sensorOUT;
        KITGripperBasisBoardIN_t* sensorIN;

        std::vector<ActorDataPtr> actorData;
        bool firstRun;
        float jumpingOffset;
        float lastAbsoluteAngle; // this is the memory for last joint angle of the PG_base joint. to detect the jumping of the sensor value

    };
    using KITGripperBasisBoardDataPtr = std::shared_ptr<KITGripperBasisBoardData>;

} // namespace armarx

