/* Jeff
 *
 */
#include "NJointTSMultiMPController.h"

namespace armarx
{
    NJointControllerRegistration<NJointTSMultiMPController> registrationControllerNJointTSMultiMPController("NJointTSMultiMPController");

    NJointTSMultiMPController::NJointTSMultiMPController(const RobotUnitPtr&, const armarx::NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&)
    {
        useSynchronizedRtRobot();
        cfg = NJointTaskSpaceDMPControllerConfigPtr::dynamicCast(config);
        ARMARX_CHECK_EXPRESSION(!cfg->nodeSetName.empty());
        VirtualRobot::RobotNodeSetPtr rns = rtGetRobot()->getRobotNodeSet(cfg->nodeSetName);

        ARMARX_CHECK_EXPRESSION_W_HINT(rns, cfg->nodeSetName);
        for (size_t i = 0; i < rns->getSize(); ++i)
        {
            std::string jointName = rns->getNode(i)->getName();
            ControlTargetBase* ct = useControlTarget(jointName, ControlModes::Velocity1DoF);
            const SensorValueBase* sv = useSensorValue(jointName);
            targets.push_back(ct->asA<ControlTarget1DoFActuatorVelocity>());

            //            const SensorValue1DoFActuatorTorque* torqueSensor = sv->asA<SensorValue1DoFActuatorTorque>();
            //            const SensorValue1DoFGravityTorque* gravityTorqueSensor = sv->asA<SensorValue1DoFGravityTorque>();

            const SensorValue1DoFActuatorVelocity* velocitySensor = sv->asA<SensorValue1DoFActuatorVelocity>();
            const SensorValue1DoFActuatorPosition* positionSensor = sv->asA<SensorValue1DoFActuatorPosition>();

            //            torqueSensors.push_back(torqueSensor);
            //            gravityTorqueSensors.push_back(gravityTorqueSensor);
            velocitySensors.push_back(velocitySensor);
            positionSensors.push_back(positionSensor);
        };

        tcp = (cfg->tcpName.empty()) ? rns->getTCP() : rtGetRobot()->getRobotNode(cfg->tcpName);
        ARMARX_CHECK_EXPRESSION_W_HINT(tcp, cfg->tcpName);

        // set tcp controller
        tcpController.reset(new CartesianVelocityController(rns, tcp));
        nodeSetName = cfg->nodeSetName;
        torquePIDs.resize(tcpController->rns->getSize(), pidController());

        ik.reset(new VirtualRobot::DifferentialIK(rns, rns->getRobot()->getRootNode(), VirtualRobot::JacobiProvider::eSVDDamped));


        finished = false;
        TaskSpaceDMPControllerConfig taskSpaceDMPConfig;
        taskSpaceDMPConfig.motionTimeDuration = cfg->timeDuration;
        taskSpaceDMPConfig.DMPKernelSize = cfg->kernelSize;
        taskSpaceDMPConfig.DMPMode = cfg->dmpMode;
        taskSpaceDMPConfig.DMPStyle = cfg->dmpStyle;
        taskSpaceDMPConfig.DMPAmplitude = cfg->dmpAmplitude;
        taskSpaceDMPConfig.phaseStopParas.goDist = cfg->phaseDist0;
        taskSpaceDMPConfig.phaseStopParas.backDist = cfg->phaseDist1;
        taskSpaceDMPConfig.phaseStopParas.Kpos = cfg->phaseKpPos;
        taskSpaceDMPConfig.phaseStopParas.Dpos = 0;
        taskSpaceDMPConfig.phaseStopParas.Kori = cfg->phaseKpOri;
        taskSpaceDMPConfig.phaseStopParas.Dori = 0;
        taskSpaceDMPConfig.phaseStopParas.mm2radi = cfg->posToOriRatio;
        taskSpaceDMPConfig.phaseStopParas.maxValue = cfg->phaseL;
        taskSpaceDMPConfig.phaseStopParas.slop = cfg->phaseK;

        taskSpaceDMPController.reset(new TaskSpaceDMPController("default", taskSpaceDMPConfig, false));

        // initialize tcp position and orientation


        NJointTSMultiMPControllerSensorData initSensorData;
        initSensorData.deltaT = 0;
        initSensorData.currentTime = 0;
        initSensorData.currentPose.setZero();
        initSensorData.currentTwist.setZero();
        controllerSensorData.reinitAllBuffers(initSensorData);



        debugName = cfg->debugName;

        KpF = cfg->Kp_LinearVel;
        KoF = cfg->Kp_AngularVel;
        DpF = cfg->Kd_LinearVel;
        DoF = cfg->Kd_AngularVel;

        filtered_qvel.setZero(targets.size());
        vel_filter_factor = cfg->vel_filter;

        filtered_position.setZero();
        pos_filter_factor = cfg->pos_filter;

        //        jlhigh = rns->getNode("..")->getJointLimitHi();
        //        jllow = rns->getNode("")->getJointLimitLo();
        firstRun = true;

        jointLowLimits.setZero(targets.size());
        jointHighLimits.setZero(targets.size());
        for (size_t i = 0; i < rns->getSize(); i++)
        {
            VirtualRobot::RobotNodePtr rn = rns->getAllRobotNodes().at(i);

            jointLowLimits(i) = rn->getJointLimitLo();
            jointHighLimits(i) = rn->getJointLimitHi();
        }

        started = false;

        NJointTSMultiMPControllerInterfaceData initInterfaceData;
        initInterfaceData.currentTcpPose = tcp->getPoseInRootFrame();
        interfaceData.reinitAllBuffers(initInterfaceData);
    }

    std::string NJointTSMultiMPController::getClassName(const Ice::Current&) const
    {
        return "NJointTSMultiMPController";
    }

    void NJointTSMultiMPController::controllerRun()
    {
        if (!started)
        {
            return;
        }

        if (!controllerSensorData.updateReadBuffer() || !taskSpaceDMPController)
        {
            return;
        }
        double deltaT = controllerSensorData.getReadBuffer().deltaT;
        Eigen::Matrix4f currentPose = controllerSensorData.getReadBuffer().currentPose;
        Eigen::VectorXf currentTwist = controllerSensorData.getReadBuffer().currentTwist;


        LockGuardType guard {controllerMutex};
        taskSpaceDMPController->flow(deltaT, currentPose, currentTwist);


        if (taskSpaceDMPController->canVal < 1e-8)
        {
            finished = true;
        }
        targetVels = taskSpaceDMPController->getTargetVelocity();
        targetPose = taskSpaceDMPController->getTargetPoseMat();
        std::vector<double> targetState = taskSpaceDMPController->getTargetPose();

        //        ARMARX_IMPORTANT << "CanVal: " << taskSpaceDMPController->canVal;
        //        ARMARX_IMPORTANT << "currentPose: " << currentPose;
        //        ARMARX_IMPORTANT << "targetVels: " << targetVels;
        //        ARMARX_IMPORTANT << "targetPose: " << targetPose;

        debugOutputData.getWriteBuffer().latestTargetVelocities["x_vel"] = targetVels(0);
        debugOutputData.getWriteBuffer().latestTargetVelocities["y_vel"] = targetVels(1);
        debugOutputData.getWriteBuffer().latestTargetVelocities["z_vel"] = targetVels(2);
        debugOutputData.getWriteBuffer().latestTargetVelocities["roll_vel"] = targetVels(3);
        debugOutputData.getWriteBuffer().latestTargetVelocities["pitch_vel"] = targetVels(4);
        debugOutputData.getWriteBuffer().latestTargetVelocities["yaw_vel"] = targetVels(5);
        debugOutputData.getWriteBuffer().dmpTargets["dmp_x"] = targetState[0];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_y"] = targetState[1];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_z"] = targetState[2];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qw"] = targetState[3];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qx"] = targetState[4];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qy"] = targetState[5];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qz"] = targetState[6];
        debugOutputData.getWriteBuffer().currentPose["currentPose_x"] = currentPose(0, 3);
        debugOutputData.getWriteBuffer().currentPose["currentPose_y"] = currentPose(1, 3);
        debugOutputData.getWriteBuffer().currentPose["currentPose_z"] = currentPose(2, 3);

        VirtualRobot::MathTools::Quaternion currentQ = VirtualRobot::MathTools::eigen4f2quat(currentPose);
        debugOutputData.getWriteBuffer().currentPose["currentPose_qw"] = currentQ.w;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qx"] = currentQ.x;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qy"] = currentQ.y;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qz"] = currentQ.z;
        debugOutputData.getWriteBuffer().currentCanVal = taskSpaceDMPController->debugData.canVal;
        debugOutputData.getWriteBuffer().mpcFactor =  taskSpaceDMPController->debugData.mpcFactor;
        debugOutputData.getWriteBuffer().error = taskSpaceDMPController->debugData.poseError;
        debugOutputData.getWriteBuffer().posError = taskSpaceDMPController->debugData.posiError;
        debugOutputData.getWriteBuffer().oriError = taskSpaceDMPController->debugData.oriError;
        debugOutputData.getWriteBuffer().deltaT = deltaT;

        debugOutputData.commitWrite();

        getWriterControlStruct().targetTSVel = targetVels;
        getWriterControlStruct().targetPose = targetPose;

        writeControlStruct();

    }


    void NJointTSMultiMPController::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        Eigen::Matrix4f currentPose = tcp->getPoseInRootFrame();
        interfaceData.getWriteBuffer().currentTcpPose = currentPose;
        interfaceData.commitWrite();

        if (firstRun)
        {
            filtered_position = currentPose.block<3, 1>(0, 3);

            firstRun = false;
            for (size_t i = 0; i < targets.size(); ++i)
            {
                targets.at(i)->velocity = 0;
            }
            return;
        }
        else
        {
            filtered_position = (1 - pos_filter_factor) * filtered_position + pos_filter_factor * currentPose.block<3, 1>(0, 3);
        }

        double deltaT = timeSinceLastIteration.toSecondsDouble();

        Eigen::MatrixXf jacobi = ik->getJacobianMatrix(tcp, VirtualRobot::IKSolver::CartesianSelection::All);

        Eigen::VectorXf qvel;
        qvel.resize(velocitySensors.size());
        Eigen::VectorXf qpos;
        qpos.resize(positionSensors.size());
        for (size_t i = 0; i < velocitySensors.size(); ++i)
        {
            qvel(i) = velocitySensors[i]->velocity;
            qpos(i) = positionSensors[i]->position;
        }

        filtered_qvel = (1 - vel_filter_factor) * filtered_qvel + vel_filter_factor * qvel;
        Eigen::VectorXf tcptwist = jacobi * filtered_qvel;

        controllerSensorData.getWriteBuffer().currentPose = currentPose;
        controllerSensorData.getWriteBuffer().currentTwist = tcptwist;
        controllerSensorData.getWriteBuffer().deltaT = deltaT;
        controllerSensorData.getWriteBuffer().currentTime += deltaT;
        controllerSensorData.commitWrite();



        Eigen::VectorXf targetVel = rtGetControlStruct().targetTSVel;
        Eigen::Matrix4f targetPose = rtGetControlStruct().targetPose;

        if (started)
        {
            targetVel = rtGetControlStruct().targetTSVel;
            targetPose = rtGetControlStruct().targetPose;
        }
        else
        {
            targetVel.setZero(6);
            targetPose = currentPose;
        }

        //        ARMARX_IMPORTANT << "targetPose: " << targetPose;

        Eigen::Matrix3f diffMat = targetPose.block<3, 3>(0, 0) * currentPose.block<3, 3>(0, 0).inverse();
        Eigen::Vector3f errorRPY = VirtualRobot::MathTools::eigen3f2rpy(diffMat);

        Eigen::VectorXf rtTargetVel;
        rtTargetVel.resize(6);
        rtTargetVel.block<3, 1>(0, 0) = KpF * (targetPose.block<3, 1>(0, 3) - filtered_position) + DpF * (targetVel.block<3, 1>(0, 0) - tcptwist.block<3, 1>(0, 0));
        rtTargetVel.block<3, 1>(3, 0) = KoF * errorRPY + DoF * (targetVel.block<3, 1>(3, 0) - tcptwist.block<3, 1>(3, 0));
        //        rtTargetVel = targetVel;



        float normLinearVelocity = rtTargetVel.block<3, 1>(0, 0).norm();
        if (normLinearVelocity > cfg->maxLinearVel)
        {
            rtTargetVel.block<3, 1>(0, 0) = cfg->maxLinearVel * rtTargetVel.block<3, 1>(0, 0) / normLinearVelocity;
        }

        float normAngularVelocity = rtTargetVel.block<3, 1>(3, 0).norm();
        if (normAngularVelocity > cfg->maxAngularVel)
        {
            rtTargetVel.block<3, 1>(3, 0) = cfg->maxAngularVel * rtTargetVel.block<3, 1>(3, 0) / normAngularVelocity;
        }


        // cartesian vel controller
        Eigen::VectorXf x;
        auto mode = rtGetControlStruct().mode;

        if (mode == VirtualRobot::IKSolver::All)
        {
            x.resize(6);
            for (size_t i = 0; i < 6; i++)
            {
                x(i) = rtTargetVel(i);
            }
        }
        else if (mode == VirtualRobot::IKSolver::Position)
        {
            x.resize(3);
            for (size_t i = 0; i < 3; i++)
            {
                x(i) = rtTargetVel(i);
            }

        }
        else if (mode == VirtualRobot::IKSolver::Orientation)
        {
            x.resize(3);
            for (size_t i = 0; i < 3; i++)
            {
                x(i) = rtTargetVel(i + 3);
            }
        }
        else
        {

            // No mode has been set
            return;
        }

        Eigen::VectorXf jnv = Eigen::VectorXf::Zero(tcpController->rns->getSize());
        float jointLimitAvoidanceKp = rtGetControlStruct().avoidJointLimitsKp;
        if (jointLimitAvoidanceKp > 0)
        {
            jnv += jointLimitAvoidanceKp * tcpController->calculateJointLimitAvoidance();
        }
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            jnv(i) += rtGetControlStruct().nullspaceJointVelocities.at(i);
        }

        Eigen::VectorXf jointTargetVelocities = tcpController->calculate(x, jnv, mode);
        ARMARX_CHECK_EXPRESSION(!targets.empty());
        ARMARX_CHECK_LESS(targets.size(), 1000);
        for (size_t i = 0; i < targets.size(); ++i)
        {
            targets.at(i)->velocity = jointTargetVelocities(i);

            if (!targets.at(i)->isValid() || fabs(targets.at(i)->velocity) > cfg->maxJointVelocity)
            {
                ARMARX_IMPORTANT << deactivateSpam(1, std::to_string(i))
                                 << "Velocity controller target is invalid - setting to zero! set value: " << targets.at(i)->velocity;
                targets.at(i)->velocity = 0.0f;
            }

        }


    }


    void NJointTSMultiMPController::learnDMPFromFiles(const Ice::StringSeq& fileNames, const Ice::Current&)
    {
        taskSpaceDMPController->learnDMPFromFiles(fileNames);

        ARMARX_INFO << "Learned DMP ... ";
    }

    void NJointTSMultiMPController::setSpeed(Ice::Double times, const Ice::Current&)
    {
        LockGuardType guard {controllerMutex};
        taskSpaceDMPController->setSpeed(times);
    }

    void NJointTSMultiMPController::setViaPoints(Ice::Double u, const Ice::DoubleSeq& viapoint, const Ice::Current&)
    {
        LockGuardType guard {controllerMutex};
        ARMARX_INFO << "setting via points ";
        taskSpaceDMPController->setViaPose(u, viapoint);
    }

    void NJointTSMultiMPController::removeAllViaPoints(const Ice::Current&)
    {
        LockGuardType guard {controllerMutex};
        ARMARX_INFO << "setting via points ";
        taskSpaceDMPController->removeAllViaPoints();
    }


    void NJointTSMultiMPController::setGoals(const Ice::DoubleSeq& goals, const Ice::Current& ice)
    {
        taskSpaceDMPController->setGoalPoseVec(goals);
    }

    void NJointTSMultiMPController::setControllerTarget(Ice::Float avoidJointLimitsKp, NJointTaskSpaceDMPControllerMode::CartesianSelection mode, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        getWriterControlStruct().avoidJointLimitsKp = avoidJointLimitsKp;
        getWriterControlStruct().mode = ModeFromIce(mode);
        writeControlStruct();
    }

    VirtualRobot::IKSolver::CartesianSelection NJointTSMultiMPController::ModeFromIce(const NJointTaskSpaceDMPControllerMode::CartesianSelection mode)
    {
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::ePosition)
        {
            return VirtualRobot::IKSolver::CartesianSelection::Position;
        }
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::eOrientation)
        {
            return VirtualRobot::IKSolver::CartesianSelection::Orientation;
        }
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::eAll)
        {
            return VirtualRobot::IKSolver::CartesianSelection::All;
        }
        ARMARX_ERROR_S << "invalid mode " << mode;
        return (VirtualRobot::IKSolver::CartesianSelection)0;
    }


    void NJointTSMultiMPController::setTorqueKp(const StringFloatDictionary& torqueKp, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            getWriterControlStruct().torqueKp.at(i) = torqueKp.at(tcpController->rns->getNode(i)->getName());
        }
        writeControlStruct();
    }

    void NJointTSMultiMPController::setNullspaceJointVelocities(const StringFloatDictionary& nullspaceJointVelocities, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            getWriterControlStruct().nullspaceJointVelocities.at(i) = nullspaceJointVelocities.at(tcpController->rns->getNode(i)->getName());
        }
        writeControlStruct();
    }

    void NJointTSMultiMPController::resetDMP(const Ice::Current&)
    {
        if (started)
        {
            ARMARX_INFO << "Cannot reset running DMP";
        }
        firstRun = true;
    }

    void NJointTSMultiMPController::stopDMP(const Ice::Current&)
    {
        mpMagazine.at(mpName).firstRun = true;
        mpMagazine.at(mpName).status = mpStatus::stopped;
        //        started = false;
        //        firstRun = true;
    }

    void NJointTSMultiMPController::runDMP(const std::string& mpName, const Ice::DoubleSeq&  goals, double tau, const Ice::Current&)
    {
        mpMagazine.at(mpName).firstRun = true;
        while (mpMagazine.at(mpName).firstRun)
        {
            usleep(100);
        }
        while (!interfaceData.updateReadBuffer())
        {
            usleep(100);
        }
        //        firstRun = true;
        //        while (firstRun)
        //        {
        //            usleep(100);
        //        }
        //        while (!interfaceData.updateReadBuffer())
        //        {
        //            usleep(100);
        //        }

        Eigen::Matrix4f pose = interfaceData.getReadBuffer().currentTcpPose;
        ARMARX_INFO << "current pose: " << pose;
        LockGuardType guard {controllerMutex};
        mpMagazine.at(mpName).mpController->prepareExecution(mpMagazine.at(mpName).mpController->eigen4f2vec(pose), goals, tau);
        mpMagazine.at(mpName).status = mpStatus::running;
        //        taskSpaceDMPController->prepareExecution(taskSpaceDMPController->eigen4f2vec(pose), goals, tau);
        //        finished = false;

        //        ARMARX_INFO << "run DMP";
        //        started = true;
    }




    void NJointTSMultiMPController::rtPreActivateController()
    {
    }

    void NJointTSMultiMPController::rtPostDeactivateController()
    {

    }

    void NJointTSMultiMPController::onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx& debugObs)
    {
        std::string datafieldName = debugName;
        StringVariantBaseMap datafields;
        auto values = debugOutputData.getUpToDateReadBuffer().latestTargetVelocities;
        for (auto& pair : values)
        {
            datafieldName = pair.first  + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        auto dmpTargets = debugOutputData.getUpToDateReadBuffer().dmpTargets;
        for (auto& pair : dmpTargets)
        {
            datafieldName = pair.first  + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        auto currentPose = debugOutputData.getUpToDateReadBuffer().currentPose;
        for (auto& pair : currentPose)
        {
            datafieldName = pair.first + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        datafieldName = "canVal_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().currentCanVal);
        datafieldName = "mpcFactor_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().mpcFactor);
        datafieldName = "error_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().error);
        datafieldName = "posError_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().posError);
        datafieldName = "oriError_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().oriError);
        datafieldName = "deltaT_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().deltaT);
        datafieldName = "DMPController_" + debugName;

        debugObs->setDebugChannel(datafieldName, datafields);
    }

    void NJointTSMultiMPController::onInitComponent()
    {
        ARMARX_INFO << "init ...";
        targetVels.resize(6);
        NJointTSMultiMPControllerControlData initData;
        initData.targetTSVel.resize(6);
        for (size_t i = 0; i < 6; ++i)
        {
            initData.targetTSVel(i) = 0;
            targetVels(i) = 0;
            initData.targetPose = tcp->getPoseInRootFrame();
        }
        initData.nullspaceJointVelocities.resize(tcpController->rns->getSize(), 0);
        initData.torqueKp.resize(tcpController->rns->getSize(), 0);
        initData.torqueKd.resize(tcpController->rns->getSize(), 0);
        initData.mode = ModeFromIce(cfg->mode);
        reinitTripleBuffer(initData);



        controllerTask = new PeriodicTask<NJointTSMultiMPController>(this, &NJointTSMultiMPController::controllerRun, 0.3);
        controllerTask->start();


    }

    void NJointTSMultiMPController::onDisconnectComponent()
    {
        ARMARX_INFO << "stopped ...";
        controllerTask->stop();
    }



}
