/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <boost/shared_ptr.hpp>

#include <VirtualRobot/Nodes/RobotNode.h>
#include <VirtualRobot/RobotNodeSet.h>

namespace armarx
{
    typedef boost::shared_ptr<class SimpleDiffIK> SimpleDiffIKPtr;

    class SimpleDiffIK
    {
    public:
        struct Parameters
        {
            Parameters() {}
            // IK params
            float ikStepLengthInitial = 0.2f;
            float ikStepLengthFineTune = 0.5f;
            size_t stepsInitial = 25;
            size_t stepsFineTune = 5;
            float maxPosError = 10.f;
            float maxOriError = 0.05f;
            float jointLimitAvoidanceKp = 2.0f;
        };
        struct Result
        {
            Eigen::VectorXf jointValues;
            Eigen::Vector3f posDiff;
            Eigen::Vector3f oriDiff;
            float posError;
            float oriError;
            bool reached;
            Eigen::VectorXf jointLimitMargins;
            float minimumJointLimitMargin;
        };
        struct Reachability
        {

            bool reachable = true;
            float minimumJointLimitMargin = M_PI;
            Eigen::VectorXf jointLimitMargins;
            float maxPosError = 0;
            float maxOriError = 0;

            void aggregate(const Result& result)
            {
                reachable = reachable && result.reached;
                minimumJointLimitMargin = std::min(minimumJointLimitMargin, result.minimumJointLimitMargin);
                if (jointLimitMargins.rows() == 0)
                {
                    jointLimitMargins = result.jointLimitMargins;
                }
                else
                {
                    jointLimitMargins = jointLimitMargins.cwiseMin(result.jointLimitMargins);
                }
                maxPosError = std::max(maxPosError, result.posError);
                maxOriError = std::max(maxOriError, result.oriError);
            }
        };

        static Result CalculateDiffIK(const Eigen::Matrix4f targetPose, VirtualRobot::RobotNodeSetPtr rns, VirtualRobot::RobotNodePtr tcp = VirtualRobot::RobotNodePtr(), Parameters params = Parameters());

        ///@brief Use this to check a trajectory of waypoints
        static Reachability CalculateReachability(const std::vector<Eigen::Matrix4f> targets, const Eigen::VectorXf& initialJV, VirtualRobot::RobotNodeSetPtr rns, VirtualRobot::RobotNodePtr tcp = VirtualRobot::RobotNodePtr(), Parameters params = Parameters());

    };
}
